unit MathParser;

interface

uses
 SysUtils, Classes, Controls, StdCtrls, ExtCtrls, Forms,
 optMathParser;

type
 TfmMathParser = class(TForm)
  pnlUp: TPanel;
   meSrc: TMemo;
   sbLayer: TScrollBox;
    imCanvas: TImage;
  meResult: TMemo;
  pnlButtons: TPanel;
   btnFxToPas: TButton;
   btnFxToFx: TButton;
   btnFxToMathExprDraw: TButton;

  procedure FormCreate(Sender: TObject);
  procedure btnFxToPasClick(Sender: TObject);
  procedure btnFxToFxClick(Sender: TObject);
  procedure btnFxToMathExprDrawClick(Sender: TObject);
 private
  p: ToptMathParser;
  Buffer: TStrings;

  procedure BuildFormulaImage;
 public
 end;

var
  fmMathParser: TfmMathParser;

implementation

uses
 Graphics,
 //   :
 QStrings, ExprDraw, ExprMake;

{$R *.dfm}

const
 FontSize = 16;

procedure TfmMathParser.FormCreate(Sender: TObject);
begin
 p := ToptMathParser.Create;
 Buffer := TStringList.Create;
end;

procedure TfmMathParser.btnFxToPasClick(Sender: TObject);
var s: string;
begin
 meResult.Clear;
 meResult.Lines.BeginUpdate;
 try
  try
   s := p.TranslateToSrcCode(10, meSrc.Text);
  except
   on E:Exception do
    begin
     meResult.Lines.Append(E.Message);
     meResult.Lines.Append(Format('  X: %d, y: %d', [p.errX, p.errY]));
    end;
  end;
 finally
  meResult.Lines.EndUpdate;
 end;

 meResult.Lines.Append(s);
 meResult.Lines.Append('');
 meResult.Lines.Append('Warnings:');
 meResult.Lines.AddStrings(p.Warnings);
end;

procedure TfmMathParser.btnFxToFxClick(Sender: TObject);
begin
 meResult.Clear;
 meResult.Lines.BeginUpdate;
 try
  try
   meSrc.Text := p.NormalizeFxText(meSrc.Text);
  except
   on E:Exception do
    begin
     meResult.Lines.Append(E.Message);
     meResult.Lines.Append(Format('  X: %d, y: %d', [p.errX, p.errY]));
    end;
  end;
 finally
  meResult.Lines.EndUpdate;
 end;
end;

procedure TfmMathParser.btnFxToMathExprDrawClick(Sender: TObject);
begin
 meResult.Clear;
 meResult.Lines.BeginUpdate;
 try
  try
   meResult.Text := p.TranslateToMathExprDraw(10, meSrc.Text);
   Buffer.Text := meResult.Text;
   BuildFormulaImage;
  except
   on E:Exception do
    begin
     meResult.Lines.Append(E.Message);
     meResult.Lines.Append(Format('  X: %d, y: %d', [p.errX, p.errY]));
    end;
  end;
 finally
  meResult.Lines.EndUpdate;
 end;
end;

procedure TfmMathParser.BuildFormulaImage;
var i, dx, dy: integer;
    tmpS: string;
    Expr: TExprClass;
    Builder: TExprBuilder;
begin
 with imCanvas.Picture.Bitmap do
  begin
   Width  := 2000;
   Height := 2000;
   Canvas.Brush.Style := bsSolid;
   Canvas.Brush.Color := sbLayer.Color;
   Canvas.FillRect(Rect(0, 0, Width, Height))
  end;

 if (Buffer.Count = 0) then Exit;

 dy := 5;
 dx := 0;

 Builder := TExprBuilder.Create;
 try
  for i := 0 to Buffer.Count - 1 do
   begin
    tmpS := Q_DelChar(Buffer[i]);

    if (tmpS <> '') then
     begin
      Expr := Builder.BuildExpr(tmpS);
      try
       Expr.Canvas := imCanvas.Picture.Bitmap.Canvas;
       Expr.Font.Size := FontSize;
       Expr.Draw(5, dy, ehLeft, evTop);
       Inc(dy, Expr.Height + 5);
       if (dx < Expr.Width) then dx := Expr.Width;
      finally
       Expr.Free;
      end;
     end;
   end;
 finally
  Builder.Free;
 end;

 dx := dx + 10;
 imCanvas.Picture.Bitmap.Width  := dx;
 imCanvas.Picture.Bitmap.Height := dy;
 imCanvas.Width  :=dx;
 imCanvas.Height :=dy
end;

end.
